"""
# mmregression
# Start 4 August 2024
# Rev 26 October 2024
"""
import arcpy
import pandas
import scipy.stats
from arcgis.features import GeoAccessor, GeoSeriesAccessor
import statsmodels.formula.api as smf
from statsmodels.stats.outliers_influence import variance_inflation_factor
def mmregression(input_features, output_features, dependent_name,
                 independent_names, standardize = True,
                transform = True, k_neighbors = 5):
    # Error checks
    if dependent_name in independent_names:
        arcpy.AddError("All independent variables should be different from the dependent variable.")
        raise arcpy.ExecuteError
    
    if len(independent_names) < 2:
        arcpy.AddError("Supply at least two independent variables.")
        raise arcpy.ExecuteError
    # Import feature class into spatial data frame
    # https://developers.arcgis.com/python/guide/introduction-to-the-spatially-enabled-dataframe/
    arcpy.AddMessage("mmregression v2024.10.26 (c) 2024 by Michael Minn")
    arcpy.AddMessage("You can redistribute or modify under the terms of")
    arcpy.AddMessage("version 2 of the GNU General Public License (GPL v2)\n\n")
    arcpy.AddMessage("Loading feature class")
                
    features = pandas.DataFrame.spatial.from_featureclass(input_features)
    features.insert(0, 'FEATUREID', range(1, len(features) + 1))
    variable_names = [dependent_name] + independent_names
    features = features[['FEATUREID'] + variable_names + ["SHAPE"]]
    features.dropna(inplace=True, ignore_index=True)
    for x in variable_names:
        features[x] = features[x].astype(float)
    if len(features) < (k_neighbors * 2):
        arcpy.AddError("There must be at least 2 * k features (n=" +
                              str(len(features)) + ", k=" + str(k_neighbors))
        raise arcpy.ExecuteError
                              
    # Transform
    arcpy.AddMessage("Transform / standardize")
    
    if transform:
        for x in variable_names:
            features[x] = 0.1 - features[x].min() + features[x]
            features[x] = scipy.stats.boxcox(list(features[x]))[0]
    # Standardize
    if standardize:
        for x in variable_names:
            features[x] = (features[x] - features[x].mean()) / features[x].std()
    # Create parallel array of nearest neighbors for each feature
    arcpy.AddMessage("Calculating nearest neighbors")
    
    features['centerx'] = features["SHAPE"].geom.centroid.apply(lambda z: z[0])
    features['centery'] = features["SHAPE"].geom.centroid.apply(lambda z: z[1])
    candidates = features[['FEATUREID','centerx', 'centery']].copy()
    neighbors = []
    for dest in range(0, len(features)):
        x1 = features.centerx[dest]
        y1 = features.centery[dest]
        candidates['distance'] = candidates.apply(lambda z: \
            (((z['centerx'] - x1)**2) +((z['centery'] - y1)**2)) ** 0.5, axis=1)
        candidates.sort_values('distance', inplace=True)
        destneighbors = candidates[candidates.distance > 0].head(n = k_neighbors)
        neighbors.append(list(destneighbors.FEATUREID))
        
    # Lagged dependent variable
    
    arcpy.AddMessage("Calculating lagged values")
    
    features['LAGVAL'] = 0.0
    for dest in range(0, len(features)):
        features.at[dest, 'LAGVAL'] = features[dependent_name] \
            [features.FEATUREID.isin(neighbors[dest])].mean()
    # Multicollinearity
    vif = pandas.DataFrame()
    vif_data = features[independent_names]
    vif['VIF'] = [variance_inflation_factor(vif_data, i) for i in range(0, len(independent_names))]
    vif['Variable'] = independent_names
    output = str(vif)
    # OLS regression
    
    arcpy.AddMessage("Running regression")
    
    formula = dependent_name + ' ~ ' + '+'.join(independent_names)
    ols_model = smf.ols(formula, data = features[variable_names]).fit()
    features['olsresiduals'] = ols_model.resid
    features['olspredicted'] = features[dependent_name] + features['olsresiduals']
    output = output + '\n\n========= OLS =========\n\n' + str(ols_model.summary())
    # Spatial lag regression
    
    formula = dependent_name + ' ~ ' + '+'.join(independent_names) + "+LAGVAL"
    lag_names = variable_names + ['LAGVAL']
    lag_model = smf.ols(formula, data = features[lag_names]).fit()
    features['lagresiduals'] = lag_model.resid
    features['lagpredicted'] = features[dependent_name] + features['lagresiduals']
    # features['lagpredicted'] = lag_model.get_prediction().predicted_mean
    # output = output + '\n\n========= Spatial Lag =========\n\n' + str(lag_model.summary())
    listing = pandas.DataFrame()
    listing["coef"] = lag_model.params.round(3)
    listing["std err"] = lag_model.bse.round(3)
    listing["t"] = lag_model.tvalues.round(3)
    listing["P>|t|"] = lag_model.pvalues.round(4)
    listing = listing.iloc[:-1 , :]
    output = output + '\n\n========= Spatial Lag =========\n\n' + listing.to_string()
    output = output + "\n\nAIC: " + str(lag_model.aic.round(4))
    output = output + "\nRho: " + str(lag_model.params['LAGVAL'].round(4))
    output = output + "\nOLS Adj R-squared: " + str(ols_model.rsquared_adj.round(3))
    # Spatial error regression needs lagged residuals from OLS
    
    features['LAGERR'] = 0.0
    
    for dest in range(0, len(features)):
        features.at[dest, 'LAGERR'] = features['olsresiduals'] \
            [features.FEATUREID.isin(neighbors[dest])].mean()
        
    formula = dependent_name + ' ~ ' + '+'.join(independent_names) + "+LAGERR"
    err_names = variable_names + ['LAGERR']
    err_model = smf.ols(formula, data = features[err_names]).fit()
    features['errresiduals'] = err_model.resid
    features['errpredicted'] = features[dependent_name] + features['errresiduals']
    # output = output + '\n\n========= Spatial Error =========\n\n' + str(err_model.summary())
    listing = pandas.DataFrame()
    listing["coef"] = err_model.params.round(3)
    listing["std err"] = err_model.bse.round(3)
    listing["t"] = err_model.tvalues.round(3)
    listing["P>|t|"] = err_model.pvalues.round(4)
    listing = listing.iloc[:-1 , :]
    output = output + '\n\n========= Spatial Error =========\n\n' + listing.to_string()
    output = output + "\n\nAIC:    " + str(err_model.aic.round(4))
    output = output + "\nLambda: " + str(err_model.params['LAGERR'].round(4))
    output = output + "\nOLS Adj R-squared: " + str(ols_model.rsquared_adj.round(3))    
    # Write output
    arcpy.AddMessage("Writing output feature class")
    features.spatial.to_featureclass(location=output_features, overwrite=True)
    
    return output
if __name__ == "__main__":
    input_features = arcpy.GetParameterAsText(0)
    output_features = arcpy.GetParameterAsText(1)
    dependent_name = arcpy.GetParameterAsText(2)
    independent_names = arcpy.GetParameterAsText(3).split(';')
    standardize = arcpy.GetParameter(4)
    transform = arcpy.GetParameter(5)
    k_neighbors = arcpy.GetParameter(6)
    results = mmregression(input_features, output_features, dependent_name,
                 independent_names, standardize,
                transform, k_neighbors = 5)
    arcpy.AddMessage(results)
    aprx = arcpy.mp.ArcGISProject('CURRENT')
    aprx.activeMap.addDataFromPath(output_features)
    
    # arcpy.SetParameterAsText(7, results)
